@@ -551,7 +551,7 @@ csi *cs_counts (const cs *A, const csi *parent, const csi *post, csi ata)
         if (parent [j] != -1) colcount [parent [j]] += colcount [j] ;
     }
     return (cs_idone (colcount, AT, w, 1)) ;    /* success: free workspace */
-} 
+}
 /* p [0..n] = cumulative sum of c [0..n-1], and then copy p [0..n-1] into c */
 double cs_cumsum (csi *p, csi *c, csi n)
 {
@@ -761,7 +761,7 @@ static csi cs_nonzero (csi i, csi j, double aij, void *other)
 csi cs_dropzeros (cs *A)
 {
     return (cs_fkeep (A, &cs_nonzero, NULL)) ;  /* keep all nonzero entries */
-} 
+}
 /* remove duplicate entries from A */
 csi cs_dupl (cs *A)
 {
@@ -1251,6 +1251,13 @@ csi *cs_maxtrans (const cs *A, csi seed)  /*[jmatch [0..m-1]; imatch [0..n-1]]*/
     for (i = 0 ; i < m ; i++) if (jmatch [i] >= 0) imatch [jmatch [i]] = i ;
     return (cs_idone (jimatch, (m2 < n2) ? C : NULL, w, 1)) ;
 }
+
+#if csi == int
+# define csi_MAX INT_MAX
+#else
+# error "Need INT_MAX analogue of  csi type"
+#endif
+
 /* C = A*B */
 cs *cs_multiply (const cs *A, const cs *B)
 {
@@ -1269,10 +1276,12 @@ cs *cs_multiply (const cs *A, const cs *B)
     Cp = C->p ;
     for (j = 0 ; j < n ; j++)
     {
-        if (nz + m > C->nzmax && !cs_sprealloc (C, 2*(C->nzmax)+m))
+        if (C->nzmax > (csi_MAX - m)/2 || // 2*C->nzmax + m  overflows
+            (nz + m > C->nzmax && !cs_sprealloc (C, 2*(C->nzmax)+m)))
         {
+            warning("Too many non-zeros in sparse product: Out of memory");
             return (cs_done (C, w, x, 0)) ;             /* out of memory */
-        } 
+        }
         Ci = C->i ; Cx = C->x ;         /* C->i and C->x may be reallocated */
         Cp [j] = nz ;                   /* column j of C starts here */
         for (p = Bp [j] ; p < Bp [j+1] ; p++)
@@ -1361,35 +1370,35 @@ csi cs_print (const cs *A, csi brief)
 {
     csi p, j, m, n, nzmax, nz, *Ap, *Ai ;
     double *Ax ;
-    if (!A) { Rprintf ("(null)\n") ; return (0) ; }
+    if (!A) { printf ("(null)\n") ; return (0) ; }
     m = A->m ; n = A->n ; Ap = A->p ; Ai = A->i ; Ax = A->x ;
     nzmax = A->nzmax ; nz = A->nz ;
-    Rprintf ("CSparse Version %d.%d.%d, %s.  %s\n", CS_VER, CS_SUBVER,
+    printf ("CSparse Version %d.%d.%d, %s.  %s\n", CS_VER, CS_SUBVER,
         CS_SUBSUB, CS_DATE, CS_COPYRIGHT) ;
     if (nz < 0)
     {
-        Rprintf ("%g-by-%g, nzmax: %g nnz: %g, 1-norm: %g\n", (double) m,
+        printf ("%g-by-%g, nzmax: %g nnz: %g, 1-norm: %g\n", (double) m,
             (double) n, (double) nzmax, (double) (Ap [n]), cs_norm (A)) ;
         for (j = 0 ; j < n ; j++)
         {
-            Rprintf ("    col %g : locations %g to %g\n", (double) j, 
+            printf ("    col %g : locations %g to %g\n", (double) j,
                 (double) (Ap [j]), (double) (Ap [j+1]-1)) ;
             for (p = Ap [j] ; p < Ap [j+1] ; p++)
             {
-                Rprintf ("      %g : %g\n", (double) (Ai [p]), Ax ? Ax [p] : 1) ;
-                if (brief && p > 20) { Rprintf ("  ...\n") ; return (1) ; }
+                printf ("      %g : %g\n", (double) (Ai [p]), Ax ? Ax [p] : 1) ;
+                if (brief && p > 20) { printf ("  ...\n") ; return (1) ; }
             }
         }
     }
     else
     {
-        Rprintf ("triplet: %g-by-%g, nzmax: %g nnz: %g\n", (double) m,
+        printf ("triplet: %g-by-%g, nzmax: %g nnz: %g\n", (double) m,
             (double) n, (double) nzmax, (double) nz) ;
         for (p = 0 ; p < nz ; p++)
         {
-            Rprintf ("    %g %g : %g\n", (double) (Ai [p]), (double) (Ap [p]),
+            printf ("    %g %g : %g\n", (double) (Ai [p]), (double) (Ap [p]),
                 Ax ? Ax [p] : 1) ;
-            if (brief && p > 20) { Rprintf ("  ...\n") ; return (1) ; }
+            if (brief && p > 20) { printf ("  ...\n") ; return (1) ; }
         }
     }
     return (1) ;
@@ -1529,6 +1538,11 @@ csi cs_qrsol (csi order, const cs *A, double *b)
 /* return a random permutation vector, the identity perm, or p = n-1:-1:0.
  * seed = -1 means p = n-1:-1:0.  seed = 0 means p = identity.  otherwise
  * p = random permutation.  */
+/*
+ * NB: We use R's RNG *and* its state; i.e., if seed is not -1 or 0,
+ * ==  'seed' is *not* used at all in this version of cs_randperm() !
+*/
+
 csi *cs_randperm (csi n, csi seed)
 {
     csi *p, k, j, t ;
